from __future__ import annotations

import ssl

from aioimaplib import IMAP4ClientProtocol, IMAP4_SSL
import asyncio

from .email_exceptions import EmailConnectionError
from python_socks.async_.asyncio import Proxy as ProxyClient
from python_socks._errors import (
    ProxyConnectionError as SocksProxyConnectionError,
    ProxyError,
    ProxyTimeoutError,
)
from python_socks import ProxyType
from typing import Optional, Callable
from better_proxy import Proxy


class ImapProxyClient(IMAP4_SSL):
    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 993,
        timeout: float = IMAP4_SSL.TIMEOUT_SECONDS,
        ssl_context: ssl.SSLContext = None,
        proxy: Optional[Proxy] = None,
    ):
        self._proxy = proxy
        self._loop = asyncio.get_running_loop()

        if not ssl_context:
            ssl_context = ssl.create_default_context()
            ssl_context.check_hostname = False
            ssl_context.verify_mode = ssl.CERT_NONE

        super().__init__(host=host, port=port, timeout=timeout, ssl_context=ssl_context)

    def create_client(
        self,
        host: str,
        port: int,
        loop: asyncio.AbstractEventLoop,
        conn_lost_cb: Callable[[Optional[Exception]], None] = None,
        ssl_context: ssl.SSLContext = None,
    ):
        self.protocol = IMAP4ClientProtocol(self._loop, conn_lost_cb)

        if ssl_context is None:
            ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)

        if self._proxy:
            self._loop.create_task(
                self._proxy_connect(
                    loop or self._loop, lambda: self.protocol, ssl_context
                )
            )
        else:
            self._loop.create_task(
                self._loop.create_connection(
                    lambda: self.protocol, host, port, ssl=ssl_context
                )
            )

    async def _proxy_connect(
        self,
        loop: asyncio.AbstractEventLoop,
        protocol_factory,
        ssl_context: ssl.SSLContext | None = None,
    ):
        proxy_type_mapping = {
            "HTTP": ProxyType.HTTP,
            "SOCKS4": ProxyType.SOCKS4,
            "SOCKS5": ProxyType.SOCKS5,
        }
        proxy_type = proxy_type_mapping.get(self._proxy.protocol, ProxyType.HTTP)

        proxy_client = ProxyClient.create(
            proxy_type=proxy_type,
            host=self._proxy.host,
            port=self._proxy.port,
            username=self._proxy.login,
            password=self._proxy.password,
            loop=loop,
        )
        try:
            sock = await proxy_client.connect(
                self.host, self.port, timeout=self.timeout
            )
            await loop.create_connection(
                protocol_factory,
                sock=sock,
                ssl=ssl_context,
                server_hostname=self.host if ssl_context else None,
            )
        except (ProxyError, SocksProxyConnectionError, ProxyTimeoutError) as e:
            raise EmailConnectionError("Proxy connection error", str(e))



from enum import StrEnum
from typing import Tuple, Optional


class EmailEncoding(StrEnum):
    UTF8 = "UTF-8"
    US_ASCII = "US-ASCII"


class EmailDomain(StrEnum):
    RAMBLER = "rambler"
    OUTLOOK = "outlook"
    FIRSTMAIL = "firstmail"
    MAILRU = "mailru"
    GMAIL = "gmail"
    CUSTOM = "custom"

    def imap_server_info(self) -> Optional[Tuple[str, EmailEncoding]]:
        """
        Retrieve the IMAP server information for the given email domain.

        Returns:
            A tuple containing the IMAP server address and the required encoding.
            Returns None if the domain does not have predefined IMAP server info.
        """
        imap_info_mapping = {
            EmailDomain.RAMBLER: ("imap.rambler.ru", EmailEncoding.UTF8),
            EmailDomain.OUTLOOK: ("outlook.office365.com", EmailEncoding.UTF8),
            EmailDomain.FIRSTMAIL: ("imap.firstmail.ltd", EmailEncoding.US_ASCII),
            EmailDomain.MAILRU: ("imap.mail.ru", EmailEncoding.UTF8),
            EmailDomain.GMAIL: ("imap.gmail.com", EmailEncoding.UTF8),
            # CUSTOM does not have predefined IMAP info
        }
        return imap_info_mapping.get(self, (None, EmailEncoding.UTF8))


def determine_email_domain(email: str) -> EmailDomain:
    """
    Determine the email domain from the provided email address.

    Args:
        email (str): The email address to evaluate.

    Returns:
        EmailDomain: The corresponding EmailDomain enum member.
    """
    try:
        str_domain = email.split("@")[1].lower()
    except IndexError:
        raise ValueError("Invalid email address format.")

    domain_mapping = {
        "rambler.ru": EmailDomain.RAMBLER,
        "outlook.com": EmailDomain.OUTLOOK,
        "mail.ru": EmailDomain.MAILRU,
        "gmail.com": EmailDomain.GMAIL,
        "firstmail.ru": EmailDomain.FIRSTMAIL,
        "senoramail.com": EmailDomain.FIRSTMAIL,
    }

    return domain_mapping.get(str_domain, EmailDomain.CUSTOM)



class BaseEmailException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class EmailConnectionError(BaseEmailException):
    def __init__(self, *args):
        super().__init__(*args)


class EmailFolderSelectionError(BaseEmailException):
    def __init__(self, *args):
        super().__init__(*args)


class EmailLoginError(BaseEmailException):
    def __init__(self, *args):
        super().__init__(*args)



from __future__ import annotations

from email import message_from_bytes
from email.utils import parsedate_to_datetime

import pytz
from better_proxy import Proxy

from better_imap.domains import determine_email_domain, EmailEncoding
from .email_exceptions import (
    EmailLoginError,
    EmailConnectionError,
    EmailFolderSelectionError,
)
from datetime import datetime, timedelta
from .imap_client import ImapProxyClient
import re
import asyncio


class MailBox:
    FOLDER_NAMES = ["INBOX", "Junk", "Spam"]

    def __init__(
        self,
        email_address: str,
        password: str,
        proxy: Proxy | None = None,
        imap_host: str | None = None,
        folder_names: list[str] = None,
        timeout: float = 30,
        encoding=EmailEncoding.UTF8,
    ):
        self._email_address = email_address
        self._password = password
        self.domain = determine_email_domain(email_address)
        self.imap_host, self.encoding = self.domain.imap_server_info()
        if not imap_host and not self.imap_host:
            raise EmailConnectionError(
                "Host not provided and domain does not have predefined imap server info"
            )

        self.imap_host = imap_host or self.imap_host
        self.encoding = encoding or self.encoding

        self._verify_data()

        self.folder_names = folder_names or self.FOLDER_NAMES

        self.email_client = ImapProxyClient(
            host=self.imap_host,
            timeout=timeout,
            proxy=proxy,
        )

    def __enter__(self):
        self._connect_to_mail()
        return self

    def __exit__(self, *args):
        self._close_connection()

    def _verify_data(self):
        if self.imap_host == "imap.rambler.ru" and "%" in self._password:
            raise Exception(
                f"IMAP password contains '%' character. Change your password"
            )

    async def _connect_to_mail(self, mailbox="INBOX"):
        try:
            await self.email_client.wait_hello_from_server()
        except Exception as e:
            raise EmailConnectionError(f"Email connection failed: {e}")
        try:
            await self.email_client.login(self._email_address, self._password)
            await self.email_client.select(mailbox=mailbox)
        except Exception as e:
            if "command SELECT illegal in state NONAUTH" in str(e):
                raise EmailLoginError(
                    f"Email account banned or login/password incorrect or IMAP not enabled: {e}"
                )

            raise EmailLoginError(f"Can not login to mail: {e}")

    async def check_email(self):
        await self._connect_to_mail()

        for mailbox in self.folder_names:
            await self._select_mailbox(mailbox)

        await self._close_connection()

    async def search_match(
        self,
        regex_pattern: str,
        from_email: str = None,
        hours_offset: int = 2,
        latest_messages: int = 10,
        start_date: datetime = None,
    ):
        if not regex_pattern:
            raise ValueError("Regex pattern must be provided to search for a match")

        if start_date is None:
            start_date = datetime.now(pytz.utc) - timedelta(hours=hours_offset)

        await self._connect_to_mail()

        matches = []

        for mailbox in self.folder_names:
            await self.email_client.select(mailbox=mailbox)
            match = await self._search_match(
                regex_pattern,
                "ALL",
                latest_messages,
                from_email,
                start_date,
            )
            if match:
                matches.append(match)

        await self._close_connection()
        return max(matches, key=lambda x: x[1])[0].strip() if matches else None

    async def search_with_retry(
        self,
        regex_pattern: str,
        from_email: str = None,
        start_date: datetime = None,
        timeout=90,
        interval=5,
        **kwargs,
    ):
        end_time = asyncio.get_event_loop().time() + timeout
        if start_date is None:
            start_date = datetime.now(pytz.utc) - timedelta(seconds=10)

        while asyncio.get_event_loop().time() < end_time:
            match = await self.search_match(
                regex_pattern=regex_pattern,
                from_email=from_email,
                start_date=start_date,
                latest_messages=5,
                **kwargs,
            )
            if match:
                return match
            await asyncio.sleep(interval)
        raise TimeoutError(f"No email received within {timeout} seconds")

    async def _search_match(
        self,
        regex_pattern: str,
        search_criteria="UNSEEN",
        latest_messages=10,
        from_email=None,
        start_date=None,
    ):

        if start_date is not None:
            date_filter = start_date.strftime("%d-%b-%Y")
            search_criteria += f" SINCE {date_filter}"

        if from_email:
            search_criteria += f' FROM "{from_email}"'

        status, data = await self.email_client.search(
            search_criteria, charset=self.encoding
        )
        if status != "OK" or not data[0]:
            return None

        email_ids = data[0].split()[-latest_messages:][::-1]
        for e_id_str in email_ids:
            email_data, message_text = await self._get_email(
                e_id_str.decode(self.encoding)
            )
            if not email_data:
                continue

            if start_date and email_data < start_date:
                continue

            if regex_pattern:
                match = self._find_first_match(regex_pattern, message_text)
                if match:
                    return match, email_data

    async def _get_email(self, email_id):
        typ, msg_data = await self.email_client.fetch(email_id, "(RFC822)")
        if typ == "OK":
            email_bytes = bytes(msg_data[1])
            email_message = message_from_bytes(email_bytes)
            email_date = parsedate_to_datetime(email_message.get("date"))

            if email_date.tzinfo is None:
                email_date = pytz.utc.localize(email_date)
            elif email_date.tzinfo != pytz.utc:
                email_date = email_date.astimezone(pytz.utc)

            message_text = self.extract_email_text(email_message)
            return email_date, message_text

    def _find_first_match(self, regex_pattern, message_text):
        matches = re.findall(regex_pattern, message_text)
        if matches:
            return matches[0]
        return None

    @staticmethod
    def extract_email_text(email_message):
        if email_message.is_multipart():
            for part in email_message.walk():
                if part.get_content_type() == "text/plain":
                    return part.get_payload(decode=True).decode("utf-8")
        msg = email_message.get_payload(decode=True)
        if msg is None:
            return ""
        return msg.decode("utf-8")

    async def _select_mailbox(self, mailbox: str):
        try:
            await self.email_client.select(mailbox=mailbox)
            if self.email_client.get_state() == "AUTH":
                raise EmailFolderSelectionError(
                    "Mail does not give access to the folder, likely IMAP is not enabled"
                )

        except TimeoutError:
            raise EmailFolderSelectionError(
                "Mail does not give access to the folder, timeout"
            )

    async def _close_connection(self):
        try:
            await self.email_client.logout()
        except Exception:
            pass
